@@ -805,6 +805,7 @@ module Sequel
       #   DB.values([[1, 2], [3, 4]]).order(:column2).limit(1, 1)
       #   # VALUES ((1, 2), (3, 4)) ORDER BY column2 LIMIT 1 OFFSET 1
       def values(v)
+        raise Error, "Cannot provide an empty array for values" if v.empty?
         @default_dataset.clone(:values=>v)
       end
 
@@ -1006,11 +1007,15 @@ module Sequel
               SQL::Function.new(:format_type, pg_type[:oid], pg_attribute[:atttypmod]).as(:db_type),
               SQL::Function.new(:pg_get_expr, pg_attrdef[:adbin], pg_class[:oid]).as(:default),
               SQL::BooleanExpression.new(:NOT, pg_attribute[:attnotnull]).as(:allow_null),
-              SQL::Function.new(:COALESCE, SQL::BooleanExpression.from_value_pairs(pg_attribute[:attnum] => SQL::Function.new(:ANY, pg_index[:indkey])), false).as(:primary_key)]}.
+              SQL::Function.new(:COALESCE, SQL::BooleanExpression.from_value_pairs(pg_attribute[:attnum] => SQL::Function.new(:ANY, pg_index[:indkey])), false).as(:primary_key),
+              Sequel[:pg_type][:typtype],
+              (~Sequel[Sequel[:elementtype][:oid]=>nil]).as(:is_array),
+            ]}.
             from(:pg_class).
             join(:pg_attribute, :attrelid=>:oid).
             join(:pg_type, :oid=>:atttypid).
             left_outer_join(Sequel[:pg_type].as(:basetype), :oid=>:typbasetype).
+            left_outer_join(Sequel[:pg_type].as(:elementtype), :typarray=>Sequel[:pg_type][:oid]).
             left_outer_join(:pg_attrdef, :adrelid=>Sequel[:pg_class][:oid], :adnum=>Sequel[:pg_attribute][:attnum]).
             left_outer_join(:pg_index, :indrelid=>Sequel[:pg_class][:oid], :indisprimary=>true).
             where{{pg_attribute[:attisdropped]=>false}}.
@@ -1537,11 +1542,12 @@ module Sequel
       end
 
       # SQL DDL statement for renaming a table. PostgreSQL doesn't allow you to change a table's schema in
-      # a rename table operation, so speciying a new schema in new_name will not have an effect.
+      # a rename table operation, so specifying a new schema in new_name will not have an effect.
       def rename_table_sql(name, new_name)
         "ALTER TABLE #{quote_schema_table(name)} RENAME TO #{quote_identifier(schema_and_table(new_name).last)}"
       end
 
+      # Handle interval and citext types.
       def schema_column_type(db_type)
         case db_type
         when /\Ainterval\z/io
@@ -1553,6 +1559,43 @@ module Sequel
         end
       end
 
+      # The schema :type entry to use for array types.
+      def schema_array_type(db_type)
+        :array
+      end
+
+      # The schema :type entry to use for row/composite types.
+      def schema_composite_type(db_type)
+        :composite
+      end
+
+      # The schema :type entry to use for enum types.
+      def schema_enum_type(db_type)
+        :enum
+      end
+
+      # The schema :type entry to use for range types.
+      def schema_range_type(db_type)
+        :range
+      end
+
+      # The schema :type entry to use for multirange types.
+      def schema_multirange_type(db_type)
+        :multirange
+      end
+
+      MIN_DATE = Date.new(-4713, 11, 24)
+      MAX_DATE = Date.new(5874897, 12, 31)
+      MIN_TIMESTAMP = Time.utc(-4713, 11, 24).freeze
+      MAX_TIMESTAMP = (Time.utc(294277) - Rational(1, 1000000)).freeze
+      TYPTYPE_METHOD_MAP = {
+        'c' => :schema_composite_type,
+        'e' => :schema_enum_type,
+        'r' => :schema_range_type,
+        'm' => :schema_multirange_type,
+      }
+      TYPTYPE_METHOD_MAP.default = :schema_column_type
+      TYPTYPE_METHOD_MAP.freeze
       # The dataset used for parsing table schemas, using the pg_* system catalogs.
       def schema_parse_table(table_name, opts)
         m = output_identifier_meth(opts[:dataset])
@@ -1568,11 +1611,33 @@ module Sequel
             row.delete(:base_oid)
             row.delete(:db_base_type)
           end
-          row[:type] = schema_column_type(row[:db_type])
+
+          db_type = row[:db_type]
+          row[:type] = if row.delete(:is_array)
+            schema_array_type(db_type)
+          else
+            send(TYPTYPE_METHOD_MAP[row.delete(:typtype)], db_type)
+          end
           identity = row.delete(:attidentity)
           if row[:primary_key]
             row[:auto_increment] = !!(row[:default] =~ /\A(?:nextval)/i) || identity == 'a' || identity == 'd'
           end
+
+          # :nocov:
+          if server_version >= 90600
+          # :nocov:
+            case row[:oid]
+            when 1082
+              row[:min_value] = MIN_DATE
+              row[:max_value] = MAX_DATE
+            when 1184, 1114
+              if Sequel.datetime_class == Time
+                row[:min_value] = MIN_TIMESTAMP
+                row[:max_value] = MAX_TIMESTAMP
+              end
+            end
+          end
+
           [m.call(row.delete(:name)), row]
         end
       end
@@ -1706,6 +1771,12 @@ module Sequel
         clone(:disable_insert_returning=>true)
       end
 
+      # Always return false when using VALUES
+      def empty?
+        return false if @opts[:values]
+        super
+      end
+
       # Return the results of an EXPLAIN query as a string
       def explain(opts=OPTS)
         with_sql((opts[:analyze] ? 'EXPLAIN ANALYZE ' : 'EXPLAIN ') + select_sql).map(:'QUERY PLAN').join("\r\n")
@@ -2125,6 +2196,11 @@ module Sequel
         "TRUNCATE TABLE#{' ONLY' if to[:only]} #{table}#{' RESTART IDENTITY' if to[:restart]}#{' CASCADE' if to[:cascade]}"
       end
 
+      # Use from_self for aggregate dataset using VALUES.
+      def aggreate_dataset_use_from_self?
+        super || @opts[:values]
+      end
+      
       # Allow truncation of multiple source tables.
       def check_truncation_allowed!
         raise(InvalidOperation, "Grouped datasets cannot be truncated") if opts[:group]
